import numpy as np
from scipy import interpolate 
from scipy import signal
from scipy.optimize import curve_fit
from scipy.interpolate import UnivariateSpline

################## SPECIFIC FUNCTIONS FOR TREATMENT OF SPECTRA

def spectrarray(name,sh,sf,x,interpmethod):
    """The function needs the x general axis, an array containing the name of files, and a precision concenrning the interpolation method
    Spectra are interpolated to have linear and constant x values and be able to be in a same array
    Interpoation methods proposed are the np.interp function or an interpolation using the interpolate 
    function of scipy, based on spline functions
    """
    for i in range(len(name)):
        rawspectre = np.genfromtxt(name[i],skip_header=sh, skip_footer=sf)
        rawspectre = rawspectre[~np.isnan(rawspectre).any(1)] # on verifie qu on a pas de nan
        # on echantillonne le spectre pour avoir les memes x
        if interpmethod == 'np.interp':
            y = np.interp(x,rawspectre[:,0],rawspectre[:,1]) # resampling 
        elif interpmethod == 'scipy.interp':
            tck = interpolate.splrep(rawspectre[:,0],rawspectre[:,1],s=0)
            y = interpolate.splev(x,tck,der=0)
        else:
            print('Error, choose the good interp method name: np.interp or scipy.interp')
        
        # Now we construct the output matrix
        # 1st column is the x axis
        # then others are the spectra in the order provided in the list of names input array
        if i == 0:
            out = np.zeros((len(x),len(name)+1))
            out[:,0]=x
            out[:,i+1]=y
        else:
            out[:,i+1]=y
            
    return out 

def spectrataux(x,spectres):
    """
    Created on Wed Feb 12 18:30:44 2014
    
    @author: Charles LE LOSQ
    Carnegie Institution of Washington D.C.
    February 2014
    This function calculates the increase/decrease rate of each frequency in a spectrum
    Your data must have the same frequency axis of course...
    Initial fitting function is a second order polynomial
    Change that following your needs
    """
    # we need an organized function before calling the curve_fit algorithm
    freq = spectres[:,0]
    # output array
    taux = np.zeros((len(freq),4));
    taux[:,0] = freq[:]
    
    # We look a each frequency, we sort y data and fit them with a second order polynomial
    for i in range(len(freq)):
        y = spectres[i,1::]
        popt, pcov = curve_fit(fun2,x,y,[0.5e-3,0.5e-4,1e-6])
        taux[i,1:len(x)]=popt
        
    return taux
    
#### OFFSETTING DATA FROM THE SPECTRARRAY FUNCTION

def spectraoffset(spectre,offsets):
    """
    This function allows to Y offset your spectra for representing them or correcting them
    with an horizontal baseline for instance
    spectre is an array constructed with the spectrarray function
    offsets is an array constructed with numpy and containing the coefficient for the offset to apply to spectre   
    """
    
    # we already say what is the output array
    out = np.zeros(spectre.shape)
    # and we offset the spectra
    for i in range(len(offsets)):
        out[:,i+1] = spectre[:,i+1] + offsets[i]
    
    return out


import numpy as np

def longcorr(data,temp,wave): # input are a two column matrix of data, temperature in C, wave as the wavelength in cm-1
    """
    # Long Correction
    # Charles Le Losq
    # CIW Washington 2014
    
    # Long's correction of Raman spectra and normalisation
    # last rev. Oct 2010, converted to Matlab and then to Python
    # ensures strictly increasing values of wavenumber
    # calc. e.s.e. as Long cor. norm. sqrt(n_raw) 3d output col.
    # exp. format to avoid null ese.
    # program long3;
    
    #Original program in pascal from J. Roux, modified for Python C. Le Losq.
    
    # See Shucker and Gammon, Phys. Rev. Lett. 1970; Galeener and Sen, Phys. Rev. B 1978; Neuville and Mysen, GCA 1996; Le Losq et al. AM 2012 and GCA 2014 for equations and theory
    """
    h = 6.62606896*10**-34   # J.s    Plank constant
    k = 1.38066e-23;     # J/K    Boltzman
    c = 2.9979e8;        # m/s    Speed of light
    v = wave;            # cm-1   Excitating laser line
    nu0 = 1.0/v*10**9;    # conversion of v in m
    T = temp + 273.15;   # C input temperature in K

    x = data[:,0]
    y = data[:,1]
    
    
    # Calculate the error on data as sqrt(y). If y <= 0, then error = sqrt of absolute value of y.
    # for doing that we construct an artificial array y containing only positive values 
    
    ese = np.sqrt(np.absolute(data[:,1])) # we assume that errors on raw data are in sqrt(y)
    
    # For retrieving the good errors after long correction, one simple way is
    # to work with relative errors...
    error = ese/y;
    
    # then we proceed to the correction (Neuville and Mysen, 1996; Le Losq et
    # al., 2012)
    
    nu = 100.0*x; # cm-1 -> m-1 Raman shift
    rnu = nu0-nu; # nu0 is in m-1
    t0 = nu0*nu0*nu0*nu/rnu/rnu/rnu/rnu;
    t1 = -h*c*nu/k/T; # c in m/s  : t1 dimensionless
    t2 = 1 - np.exp(t1);
    longsp = y*t0*t2; # for y values
    #long2 = ese*t0*t2; # for errors, as comment as we use relative errors
    
    
    # normalized to max intensity
    # tried area with np.trapz but their is an issue for now
    norm = np.trapz(longsp,x)
    #norm = np.max(longsp)
    longsp = longsp/norm
    
    eselong = error*longsp
    
    spectreout = np.zeros((len(x),3))
    spectreout[:,0] = x
    spectreout[:,1] = longsp
    spectreout[:,2] = eselong
    
    return spectreout


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 12 18:30:44 2014

@author: Charles LE LOSQ
Carnegie Institution of Washington D.C.
February 2014
This function calculates the increase/decrease rate of each frequency in a spectrum
Your data must have the same frequency axis of course...
Initial fitting function is a second order polynomial
Change that following your needs
"""
import numpy as np
from scipy.optimize import curve_fit

def fun(x,a,b,c):
        y = a + b*x + c*x*x
        return y

def taux(x,spectres):
    
    # we need an organized function before calling the curve_fit algorithm
    freq = spectres[:,0]
    # output array
    taux = np.zeros((len(freq),4));
    taux[:,0] = freq
    
    # We look a each frequency, we sort y data and fit them with a second order polynomial
    for i in range(len(freq)):
        y = spectres[i,1::]
        popt, pcov = curve_fit(fun,x,y,[0.5e-3,0.5e-4,1e-6])
        taux[i,1:len(x)]=popt
        
    return taux




import numpy as np
from gcvspline import gcvspline, splderivative 
from scipy.optimize import curve_fit
from scipy.interpolate import UnivariateSpline
from scipy.spatial import ConvexHull
import scipy.sparse as sparse
from numpy.linalg import norm
from sklearn import preprocessing

def get_portion_interest(x,y,bir):
    """
    This function extracts the signals indicated in the bir.
    
    Inputs
    ------
    
    x: the x axis
    
    y: the y values
    
    bir: the regions x values where the signal needs to be extracted, must be a n x 2 dimension array
    
    Outputs
    -------
    
    an 2 columns x-y array containing the signals in the bir.
    
    """
    birlen = np.array(bir.shape[0])
    
    sp = np.transpose(np.vstack((x,y)))
    ### selection of bir data
    for i in range(birlen):
        if i == 0:
            yafit = sp[np.where((sp[:,0]> bir[i,0]) & (sp[:,0] < bir[i,1]))] 
        else:
            je = sp[np.where((sp[:,0]> bir[i,0]) & (sp[:,0] < bir[i,1]))] 
            yafit = np.vstack((yafit,je))
            
    return yafit
    

def baseline(x_input,y_input,bir,method, **kwargs):
    """
    This function allows subtracting a baseline under the spectra
    spectre is a spectrum or an array of spectra constructed with the spectrarray function
    bir contains the Background Interpolation Regions, it must be a n x 2 dimension array
    
    Inputs
    ------
    
        X: Array with x values.
    
        Y: array with y values.
    
        bir: an Array containing the regions of interest, organised per line. for instance, roi = np.array([[100., 200.],[500.,600.]]) will define roi between 100 and 200 as well as between 500 and 600,.
    
        methods:
    
            "poly": polynomial fitting, with splinesmooth the degree of the polynomial.
    
            "unispline": spline with the UnivariateSpline function of Scipy, splinesmooth is the spline smoothing factor (assume equal weight in the present case);
    
            "gcvspline": spline with the gcvspl.f algorythm, really robust. Spectra must have x, y, ese in it, and splinesmooth is the smoothing factor;
            for gcvspline, if ese are not provided we assume ese = sqrt(y);
    
            'exp': baseline of type 
    
            'log': baseline of type a*np.log(-b*(x-c))-d*x**2, parameters adjusted by scipy.optimize.curve_fit
    
            "rubberband": rubberband baseline fitting
    
            "als": automatic least square fitting following Eilers and Boelens 2005.
    
            "arPLS": automatic baseline fit using the algorithm from Baek et al. 2015 Baseline correction using asymmetrically reweighted penalized least suqares smoothing, Analyst 140: 250-257.
    
    Options
    -------
    
        Options for the ALS algorithm are:
    
        polynomial_order: integer, the degree of the polynomial (0 for a constant), default = 1
        
        s: float, spline smoothing coefficient for the unispline and gcvspline algorithms
    
        lam: float, between 10**2 to 10**9, default = 10**5
    
        p: float, advised value between 0.001 to 0.1, default = 0.01
        niter: number of iteration, default = 10
    
        Options for the ALS algorithm are:
    
        lam: float, the lambda smoothness parameter. Typical values are between 10**2 to 10**9, default = 10**5
    
        ratio = float, the termination ratio. Set it high to start and lower it. Default = 0.01.
    
    Outputs
    -------
    
        out1: an 2 columns x-y array containing the corrected signal
    
        out2: an 2 columns x-y array containing the baseline
    
        coefs: contains spline coefficients.
    
    """
    # we get the signals in the bir
    yafit_unscaled = get_portion_interest(x_input,y_input,bir)
     
    # signal standard standardization with sklearn
    # this helps for polynomial fitting
    X_scaler = preprocessing.StandardScaler().fit(x_input.reshape(-1, 1))
    Y_scaler = preprocessing.StandardScaler().fit(y_input.reshape(-1, 1))
    
    # transformation
    x = X_scaler.transform(x_input.reshape(-1, 1))
    y = Y_scaler.transform(y_input.reshape(-1, 1))
    
    print(x.shape)         
    
    yafit = np.copy(yafit_unscaled)
    yafit[:,0] = X_scaler.transform(yafit_unscaled[:,0].reshape(-1, 1))[:,0]
    yafit[:,1] = Y_scaler.transform(yafit_unscaled[:,1].reshape(-1, 1))[:,0]

    y = y.reshape(len(y_input))
    
    if method == 'poly':
        
        # optional parameters
        poly_order = kwargs.get('polynomial_order',1)
        
        coeffs = np.polyfit(yafit[:,0],yafit[:,1],poly_order)
                
        baseline_fitted = np.polyval(coeffs,x)
        y_corrected = y-baseline_fitted.reshape(-1)
        
    elif method == 'unispline':
        
        # optional parameters
        splinesmooth = kwargs.get('s',2.0)
        
        # fit of the baseline
        coeffs = UnivariateSpline(yafit[:,0],yafit[:,1], s=splinesmooth)
            
        baseline_fitted = coeffs(x)
        y_corrected = y-baseline_fitted.reshape(-1)
        
    elif method == 'gcvspline':
        
        # optional parameters
        splinesmooth = kwargs.get('s',2.0)
        
        # Spline baseline with mode 1 of gcvspl.f, see gcvspline documentation
        c, wk, ier = gcvspline(yafit[:,0],yafit[:,1],np.sqrt(np.abs(yafit[:,1])),splinesmooth,splmode = 1) # gcvspl with mode 1 and smooth factor
        
        baseline_fitted = splderivative(x,yafit[:,0],c)       
        y_corrected = y-baseline_fitted.reshape(-1)
            
    elif method == 'exp':
        ### Baseline is of the type y = a*exp(b*(x-xo))
        # optional parameters
        p0_exp = kwargs.get('p0_exp',[1.,1.,1.])
        ## fit of the baseline
        coeffs, pcov = curve_fit(funexp,yafit[:,0],yafit[:,1],p0 = p0_exp)
        
        baseline_fitted = funexp(x,coeffs[0],coeffs[1],coeffs[2])
        y_corrected = y-baseline_fitted.reshape(-1)
    
    elif method == 'log':
        ### Baseline is of the type y = a*exp(b*(x-xo))
        # optional parameters
        p0_log = kwargs.get('p0_log',[1.,1.,1.,1.])
        ## fit of the baseline
        coeffs, pcov = curve_fit(funlog,yafit[:,0],yafit[:,1],p0 = p0_exp)
        
        baseline_fitted = funlog(x,coeffs[0],coeffs[1],coeffs[2],coeffs[3])
        y_corrected = y-baseline_fitted.reshape(-1)
            
    elif method == 'rubberband':
        # code from this stack-exchange forum
        #https://dsp.stackexchange.com/questions/2725/how-to-perform-a-rubberband-correction-on-spectroscopic-data
        
        # Find the convex hull
        v = ConvexHull(np.array([x, y])).vertices
            
        # Rotate convex hull vertices until they start from the lowest one
        v = np.roll(v, -v.argmin())
        # Leave only the ascending part
        v = v[:v.argmax()]

        # Create baseline using linear interpolation between vertices
        baseline_fitted = np.interp(x, x[v], y[v])
        y_corrected = y-baseline_fitted
        
    elif method == 'als':
        # Matlab code in Eilers et Boelens 2005
        # Python addaptation found on stackoverflow: https://stackoverflow.com/questions/29156532/python-baseline-correction-library
        
        # optional parameters
        lam = kwargs.get('lam',1.0*10**5)
        p = kwargs.get('p',0.01)
        niter = kwargs.get('niter',10)
        
        # starting the algorithm
        L = len(y)
        D = sparse.csc_matrix(np.diff(np.eye(L), 2))
        w = np.ones(L)
        for i in range(niter):
            W = sparse.spdiags(w, 0, L, L)
            Z = W + lam * D.dot(D.transpose())
            z = sparse.linalg.spsolve(Z, w*y)
            w = p * (y > z) + (1-p) * (y < z)
        
        baseline_fitted = z
        y_corrected = y-baseline_fitted
        
    elif method == 'arPLS':
        # Adaptation of the Matlab code in Baek et al 2015 
        
        # optional parameters
        lam = kwargs.get('lam',1.0*10**5)
        ratio = kwargs.get('ratio',0.01)
        
        N = len(y)
        D = sparse.csc_matrix(np.diff(np.eye(N), 2))
        w = np.ones(N)

        while True:
            W = sparse.spdiags(w, 0, N, N)
            Z = W + lam * D.dot(D.transpose())
            z = sparse.linalg.spsolve(Z, w*y)
            d = y - z
            # make d- and get w^t with m and s
            dn = d[d<0]
            m = np.mean(dn)
            s = np.std(dn)
            wt = 1.0/(1 + np.exp( 2* (d-(2*s-m))/s ) )
            # check exit condition and backup
            if norm(w-wt)/norm(w) < ratio:
                break
            w = wt

        baseline_fitted = z
        y_corrected = y-baseline_fitted

    return Y_scaler.inverse_transform(y_corrected.reshape(-1, 1)), Y_scaler.inverse_transform(baseline_fitted.reshape(-1, 1))
    #return y_corrected, baseline_fitted
    
    #


import numpy as np
from scipy import signal

def smooth(x,window_len=11,window="hamming"):
    """smooth the data using a window with requested size.
    
    Taken from http://wiki.scipy.org/Cookbook/SignalSmooth
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")


    s=np.r_[x[window_len-1:0:-1],x,x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=np.ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')

    y=np.convolve(w/w.sum(),s,mode='valid')
    return y[(window_len/2-1):-(window_len/2)] #Not working pretty well....
    

def whittaker(x,**kwargs):
    """
    Whittaker smoother from Eilers 2003
    
    Inputs
    ------
    
    x: vector with the values to smooth (equally spaced)
    
    Options
    -------
    
    Lambda: the smoothing coefficient, the higher the smoother. Default = 10^5.
    
    Outputs
    -------
    
    A vector containing the smoothed values
    
    """
    # optional parameters
    lam = kwargs.get('Lambda',1.0*10**5)
    
    # starting the algorithm
    L = len(y)
    D = sparse.csc_matrix(np.diff(np.eye(L), 2))
    w = np.ones(L)
    W = sparse.spdiags(w, 0, L, L)
    Z = W + lam * D.dot(D.transpose())
    z = sparse.linalg.spsolve(Z, w*y)
    
    return z

#### FILTERING OF DATA WITH A BUTTERWORTH FILTER 
   
def spectrafilter(spectre,filtertype,fq,numtaps,columns):
    """
    This function filters the HF of the spectra
    It use a low pass butterworth filter
    y is an array of spectra constructed with the previous spectraarray function
    filtertype contains the string defining which type of filter you want
    Choose between low, high, bandstop, bandpass
    fq is the frequency of the periodic signal you try to erase
    if it is a bandpass or band stop, f must be an array containing the cutoff frequencies
    columns is an array defining which columns you want to treat
    first column of y must contain the frequency
    """
    
    # we already say what is the output array
    out = np.zeros(spectre.shape)
    
    # Butterworth band stop filter caracteristics
    a = spectre[1,0] - spectre[0,0]
    samplerate = 1/a  #Hertz
    nyq_rate = samplerate/2 # frequence Nyquist
    cutf = fq # cutoff frequency
    #bandwidth = 0.005 # largeur filtre, for band pass/stop filters
    numtaps = 1 # ordre du filtre...
    
    for i in range(len(columns)):
        y = spectre[:,columns[i]]
        if (filtertype == 'low') or (filtertype == 'high'):
            b, a = signal.butter(numtaps, [(cutf/nyq_rate)], btype = filtertype)
            out[:,columns[i]] = signal.filtfilt(b, a, y) # filter with phase shift correction
        else:
            b, a = signal.butter(numtaps, [(cutf[0]/nyq_rate),(cutf[1]/nyq_rate)], btype = filtertype)
            out[:,columns[i]] = signal.filtfilt(b, a, y) # filter with phase shift correction

    # Note forgetting to register the x axis        
    out[:,0] = spectre[:,0]
    
    return out


from numpy.distutils.core import setup
from numpy.distutils.extension import Extension

setup(name='rampy',
      version='0.2.5',
      description='A Python module containing functions to treat spectroscopic (XANES, Raman, IR...) data',
      url='https://github.com/charlesll/rampy',
      author='Charles Le Losq',
      author_email='charles.lelosq@anu.edu.au',
      license='GNU-GPLv2',
      packages=['rampy', 'rampy.tests'],
      install_requires=['numpy>=1.12','gcvspline','scipy','sklearn'],
      zip_safe=False)



import unittest

import sys
sys.path.append("../../")

import numpy as np
import scipy

import gcvspline, rampy

class TestBaseline(unittest.TestCase):
    
    def test_baseline(self):

        x2 = np.arange(1,100,0.5)
        y2 = 2*x2 + 20.0 * np.exp(-np.log(2) * ((x2-50.0)/10.0)**2)

        # need to define some fitting regions for the spline
        roi2 = np.array([[1,30],[80,100]])

        # calculating the baselines
        ycalc1, base1 = rampy.baseline(x2,y2,roi2,'poly',polynomial_order=1)
        ycalc2, base2 = rampy.baseline(x2,y2,roi2,'gcvspline',s=0.1 )
        ycalc3, base3 = rampy.baseline(x2,y2,roi2,'unispline',s=1e0)
        ycalc4, base4 = rampy.baseline(x2,y2,roi2,'als',lam=10**7,p=0.05)
        ycalc5, base5 = rampy.baseline(x2,y2,roi2,'arPLS',lam=10**7,ratio=0.1)
        
        np.testing.assert_equal(ycalc1.shape,base1.shape)
        
        np.testing.assert_almost_equal(2*x2,base1[:,0],0)
        np.testing.assert_almost_equal(2*x2,base2[:,0],0)
        np.testing.assert_almost_equal(2*x2,base3[:,0],0)
        np.testing.assert_almost_equal(2*x2,base4[:,0],0)
        np.testing.assert_almost_equal(2*x2,base5[:,0],0)
        
if __name__ == '__main__':
    unittest.main()




